from llvmlite import ir
from numba import types
from numba.core import cgutils, utils
from numba.cpython.hashing import _Py_hash_t
from numba.extending import (
    as_numba_type,
    box,
    intrinsic,
    lower_builtin,
    make_attribute_wrapper,
    models,
    overload_attribute,
    overload_method,
    register_model,
    type_callable,
    typeof_impl,
)

from pybond import Bond
from pybond.ffi import (
    bond_accrued_interest,
    bond_calc_ytm,
    bond_clean_price,
    bond_coupon_rate,
    bond_dirty_price,
    bond_duration,
)
from pybond.nb.nb_datetime import DateTimeType  # , create_bond

from .nb_date import DateType


class BondType(types.Type):
    def __init__(self):
        super().__init__(name="Bond")


bond_type = BondType()
as_numba_type.register(Bond, bond_type)


@typeof_impl.register(Bond)
def typeof_bond(val, c):
    return bond_type


@type_callable(Bond)
def type_bond(context):
    def typer(val):
        return bond_type

    return typer


@register_model(BondType)
class BondModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("ptr", types.voidptr),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


make_attribute_wrapper(BondType, "ptr", "ptr")


@lower_builtin(Bond, types.string)
def impl_bond_builder(context, builder, sig, args):
    typ = sig.return_type
    (val,) = args
    # Get string data from Numba string
    code = context.make_helper(builder, types.string, val)
    fn = cgutils.get_or_insert_function(
        builder.module,
        ir.FunctionType(
            ir.PointerType(ir.IntType(8)),
            [ir.PointerType(ir.IntType(8)), ir.IntType(utils.MACHINE_BITS)],
        ),
        name="create_bond",
    )
    ptr = builder.call(fn, [code.data, code.length])
    # Create Bond object
    bond = cgutils.create_struct_proxy(typ)(context, builder)
    bond.ptr = ptr
    return bond._getvalue()


@overload_attribute(BondType, "coupon_rate")
def bond_attr_coupon_rate(bond):
    def impl(bond):
        return bond_coupon_rate(bond.ptr)

    return impl


def ir_get_bond_full_code(ptr, context, builder):
    """根据Bond的指针获取bond的代码(包括交易所信息)"""
    fn = cgutils.get_or_insert_function(
        builder.module,
        ir.FunctionType(ir.PointerType(ir.IntType(8)), [ir.PointerType(ir.IntType(8))]),
        name="bond_full_code",
    )
    cstr = builder.call(fn, [ptr])
    # 使用 LLVM strlen
    strlen_fn = cgutils.get_or_insert_function(
        builder.module,
        ir.FunctionType(ir.IntType(64), [ir.PointerType(ir.IntType(8))]),
        name="strlen",
    )
    length = builder.call(strlen_fn, [cstr])
    uni_str = cgutils.create_struct_proxy(types.unicode_type)(context, builder)
    uni_str.data = cstr
    uni_str.length = length
    uni_str.kind = ir.Constant(ir.IntType(32), 1)  # kind=1 is PY_UNICODE_1BYTE_KIND
    uni_str.is_ascii = ir.Constant(ir.IntType(32), 1)
    uni_str.meminfo = context.get_constant_null(types.voidptr)
    # Set hash to -1 to indicate that it should be computed.
    # We cannot bake in the hash value because of hashseed randomization.
    uni_str.hash = context.get_constant(_Py_hash_t, -1)
    uni_str.parent = cgutils.get_null_value(uni_str.parent.type)
    return uni_str._getvalue()


@intrinsic
def get_bond_full_code(typingctx, bond_ptr):
    """根据Bond的指针获取bond的代码(包括交易所信息)"""

    def codegen(context, builder, sig, args):
        return ir_get_bond_full_code(args[0], context, builder)

    sig = types.unicode_type(bond_ptr)
    return sig, codegen


@overload_attribute(BondType, "full_code")
def get_full_code_attr(bond):
    if isinstance(bond, BondType):

        def impl(bond):
            return get_bond_full_code(bond.ptr)

        return impl


@overload_method(BondType, "duration")
def bond_calc_duration(bond, ytm, date):
    if not isinstance(date, (DateType, DateTimeType)):
        return

    def impl(bond, ytm, date):
        return bond_duration(bond.ptr, ytm, date.year, date.month, date.day)

    return impl


@overload_method(BondType, "accrued_interest")
def bond_calc_accrued_interest(bond, date):
    if not isinstance(date, (DateType, DateTimeType)):
        return

    def impl(bond, date):
        return bond_accrued_interest(bond.ptr, date.year, date.month, date.day)

    return impl


@overload_method(BondType, "dirty_price")
def bond_calc_dirty_price(bond, ytm, date):
    if not isinstance(date, (DateType, DateTimeType)):
        return

    def impl(bond, ytm, date):
        return bond_dirty_price(bond.ptr, ytm, date.year, date.month, date.day)

    return impl


@overload_method(BondType, "clean_price")
def bond_calc_clean_price(bond, ytm, date):
    if not isinstance(date, (DateType, DateTimeType)):
        return

    def impl(bond, ytm, date):
        return bond_clean_price(bond.ptr, ytm, date.year, date.month, date.day)

    return impl


@overload_method(BondType, "calc_ytm_with_price")
def bond_ytm_with_price(bond, dirty_price, date):
    if not isinstance(date, (DateType, DateTimeType)):
        return

    def impl(bond, dirty_price, date):
        return bond_calc_ytm(bond.ptr, dirty_price, date.year, date.month, date.day)

    return impl


@box(BondType)
def box_bond(typ, val, c):
    """
    Convert a native Bond structure to python Bond.
    """
    bond = cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
    bond_code = ir_get_bond_full_code(bond.ptr, c.context, c.builder)
    # Call Bond new to create a new Bond object
    val_obj = c.pyapi.from_native_value(types.unicode_type, bond_code)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(Bond))
    res = c.pyapi.call_function_objargs(class_obj, (val_obj,))
    c.pyapi.decref(val_obj)
    c.pyapi.decref(class_obj)
    return res



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import polars as pl

if TYPE_CHECKING:
    from polars._typing import IntoExpr

from .polars_utils import parse_into_expr, register_plugin


class Fee(pl.Expr):
    def __new__(cls, fee_dict=None):
        if fee_dict is None:
            fee_dict = {"kind": "zero"}
        obj = super().__new__(cls)
        obj._fee_dict = fee_dict
        return obj

    @property
    def kind(self):
        return self._fee_dict.get("kind")

    @property
    def fee(self):
        return self._fee_dict.get("fee")

    @property
    def _pyexpr(self):
        return pl.lit(json.dumps(self._fee_dict))._pyexpr

    def __add__(self, other: Fee | int | float):
        left_items = self._fee_dict["items"] if self.kind == "sum" else [self._fee_dict]
        if isinstance(other, Fee):
            right_items = (
                other._fee_dict["items"] if other.kind == "sum" else [other._fee_dict]
            )
            return Fee({"kind": "sum", "items": left_items + right_items})
        elif isinstance(other, (int, float)) and self.kind in [
            "fixed",
            "per_trade",
            "per_qty",
            "percent",
        ]:
            _dict = self._fee_dict.copy()
            _dict["fee"] += other
            return Fee(_dict)
        else:
            msg = f"Cannot add {other} on {self.kind} Fee"
            raise NotImplementedError(msg)

    def __radd__(self, other):
        return self.__add__(other)

    def min(self, fee: float):
        """
        fee = min(cap, inner_fee)
        """
        return Fee({"kind": "min", "fee": self._fee_dict, "cap": fee})

    def max(self, fee: float):
        """
        fee = max(floor, inner_fee)
        """
        return Fee({"kind": "max", "fee": self._fee_dict, "floor": fee})

    @staticmethod
    def zero():
        """Represents a fee of zero."""
        return Fee({"kind": "zero"})

    @staticmethod
    def fixed(fee: float) -> TradeFee:
        """Represents a fixed fee for a trade."""
        return Fee({"kind": "fixed", "fee": fee})

    @staticmethod
    def trade(fee: float) -> TradeFee:
        """Represents a fixed fee for a trade."""
        return Fee({"kind": "per_trade", "fee": fee})

    @staticmethod
    def qty(fee: float) -> QtyFee:
        """Represents a fee based on the quantity of a trade."""
        return Fee({"kind": "per_qty", "fee": fee})

    @staticmethod
    def percent(fee: float) -> PercentFee:
        """Represents a fee based on a percentage of the trade amount."""
        return Fee({"kind": "percent", "fee": fee})


class FeeZero(Fee):
    def __new__(cls):
        obj = super().__new__(cls)
        obj._fee_dict = Fee.zero()._fee_dict
        return obj


class PercentFee(Fee):
    """Represents a fee based on a percentage of the trade amount."""

    def __new__(cls, rate: float):
        obj = super().__new__(cls)
        obj._fee_dict = Fee.percent(rate)._fee_dict
        return obj


class QtyFee(Fee):
    """Represents a fee based on the quantity of a trade."""

    def __new__(cls, per_qty: float):
        obj = super().__new__(cls)
        obj._fee_dict = Fee.qty(per_qty)._fee_dict
        return obj


class TradeFee(Fee):
    """Represents a fixed fee for a trade."""

    def __new__(cls, per_trade: float):
        obj = super().__new__(cls)
        obj._fee_dict = Fee.trade(per_trade)._fee_dict
        return obj


def calc_bond_trade_pnl(
    symbol: IntoExpr,
    settle_time: IntoExpr,
    qty: IntoExpr | None = None,
    clean_price: IntoExpr | None = None,
    clean_close: IntoExpr = "close",
    bond_info_path: str | None = None,
    multiplier: IntoExpr | None = None,
    fee: IntoExpr | Fee | None = None,
    begin_state: IntoExpr | None = None,
) -> pl.Expr:
    """
    计算债券交易pnl
    symbol: 交易的标的名称, 如果不是债券传⼊空字符串即可。
    settle_time: 结算时间,
        如果settle_time传入代表Trade的struct Series(包含time, price, qty三个field), 则可以不传qty和clean_price
    qty: 成交量, 正负号表⽰⽅向
    clean_price: 成交的净价
    clean_close: 当前时间段的最新价格(净价)
    bond_info_path: 可以指定债券信息的存放⽂件夹, 不传⼊则使⽤默认路径.
    multiplier: 合约乘数, 例如对于债券, 1000的成交对应1000w, 合约乘数应为100, 默认为1
    fee: 交易费⽤
    费⽤设置说明:
        TradeFee: 每笔成交⽀付的费⽤
        QtyFee: 每⼿需要⽀付的费⽤
        PercentFee: 按照成交⾦额百分⽐⽀付的费⽤
        费⽤⽀持相加, 例如 QtyFee(120) + TradeFee(20)
    """
    assert clean_close is not None
    if fee is None:
        fee = Fee.zero()
    fee = parse_into_expr(fee)
    symbol = parse_into_expr(symbol)
    settle_time = parse_into_expr(settle_time)
    clean_close = parse_into_expr(clean_close)
    multiplier = parse_into_expr(multiplier)
    if begin_state is not None and not isinstance(begin_state, dict):
        begin_state = parse_into_expr(begin_state)
    if bond_info_path is None:
        from .bond import bonds_info_path as path

        bond_info_path = str(path)

    if begin_state is None:
        begin_state = pl.lit(
            {
                "pos": 0,
                "avg_price": 0,
                "pnl": 0,
                "realized_pnl": 0,
                "pos_price": 0,
                "unrealized_pnl": 0,
                "coupon_paid": 0,
                "amt": 0,
                "fee": 0,
            }
        )
    kwargs = {"bond_info_path": bond_info_path}
    if all(x is None for x in [qty, clean_price]):
        # struct settle_time, contains trade info
        args = [symbol, settle_time, clean_close, begin_state, multiplier, fee]
    else:
        qty = parse_into_expr(qty)
        clean_price = parse_into_expr(clean_price)
        args = [
            symbol,
            settle_time,
            qty,
            clean_price,
            clean_close,
            begin_state,
            multiplier,
            fee,
        ]
    return register_plugin(
        args=args,
        kwargs=kwargs,
        symbol="calc_bond_trade_pnl",
        is_elementwise=False,
    )


def calc_trade_pnl(
    time: IntoExpr,
    qty: IntoExpr | None = None,
    price: IntoExpr | None = None,
    close: IntoExpr = "close",
    multiplier: IntoExpr | None = None,
    fee: IntoExpr | Fee | None = None,
    begin_state: IntoExpr | None = None,
):
    """
    计算交易pnl
    symbol: 交易的标的名称, 如果不是债券传⼊空字符串即可。
    time: 交易时间,
        如果time传入代表Trade的struct Series(包含time, price, qty三个field), 则可以不传qty和clean_price
    qty: 成交量, 正负号表⽰⽅向
    clean_price: 成交的净价
    clean_close: 当前时间段的最新价格(净价)
    multiplier: 合约乘数, 例如对于债券, 1000的成交对应1000w, 合约乘数应为100, 默认为1
    fee: 交易费⽤
    费⽤设置说明:
        TradeFee: 每笔成交⽀付的费⽤
        QtyFee: 每⼿需要⽀付的费⽤
        PercentFee: 按照成交⾦额百分⽐⽀付的费⽤
        费⽤⽀持相加, 例如 QtyFee(120) + TradeFee(20)
    """
    return calc_bond_trade_pnl(
        symbol=pl.lit(""),
        settle_time=time,
        qty=qty,
        clean_price=price,
        clean_close=close,
        multiplier=multiplier,
        fee=fee,
        begin_state=begin_state,
    )


def trading_from_pos(
    time: IntoExpr,
    pos: IntoExpr,
    open: IntoExpr,
    finish_price: IntoExpr | None = None,
    cash: IntoExpr = 1e8,
    multiplier: IntoExpr | None = None,
    qty_tick: IntoExpr = 1.0,
    min_adjust_amt: float = 0.0,
    *,
    stop_on_finish: bool = False,
    keep_shape: bool = False,
) -> pl.Expr:
    """
    生成交易记录
    time: ⽤于⽣成成交时间, ⽀持任意可以转为polars表达式的输⼊
    pos: 当前时间的实际仓位, -1 ~ 1, 表⽰百分⽐
    open: 当前周期的开仓价格
    cash: 总资⾦, ⽤于计算实际开仓⼿数
    multiplier: 合约乘数, 默认为1
    qty_tick: 最⼩开仓⼿数, 例如0.01, 0.1, 1, 100
    stop_on_finish: 当前标的没有数据后是否平仓
    finish_price: 当前标的没数据时的平仓价格, ⽀持polars表达式
    keep_shape: 是否维持表达式的长度, 不保留则只返回实际发生的交易
    """
    time = parse_into_expr(time)
    pos = parse_into_expr(pos)
    open = parse_into_expr(open)
    multiplier = parse_into_expr(multiplier)
    qty_tick = parse_into_expr(qty_tick)
    if finish_price is not None:
        stop_on_finish = True
    finish_price = parse_into_expr(finish_price)
    cash = parse_into_expr(cash)
    kwargs = {
        "cash": None,  # 会从表达式中获取
        "multiplier": 0.0,  # 会从表达式中获取
        "qty_tick": 1.0,  # 会从表达式中获取
        "stop_on_finish": stop_on_finish,
        "finish_price": None,  # 会从表达式中获取
        "min_adjust_amt": float(min_adjust_amt),
        "keep_shape": bool(keep_shape),
    }
    return register_plugin(
        args=[time, pos, open, finish_price, cash, multiplier, qty_tick],
        kwargs=kwargs,
        symbol="trading_from_pos",
        is_elementwise=False,
    )



# from datetime import date, datetime, time

from .nb_bond import Bond, BondType, bond_type
from .nb_date import DateType, date_type
from .nb_datetime import DateTime, DateTimeType, datetime_type
from .nb_duration import Duration, DurationType, duration_type
from .nb_evaluators import TfEvaluator, TfEvaluatorType, tf_evaluator_type
from .nb_time import Time, TimeType, time_type

__all__ = [
    "Bond",
    "BondType",
    "DateTime",
    "DateTimeType",
    "DateType",
    "Duration",
    "DurationType",
    "TfEvaluator",
    "TfEvaluatorType",
    "Time",
    "TimeType",
    "bond_type",
    # "date",
    "date_type",
    # "datetime",
    "datetime_type",
    "duration_type",
    "tf_evaluator_type",
    # "time",
    "time_type",
]



import ctypes

from .lib import lib

build_datetime_ns = lib.build_datetime_ns
build_datetime_ns.argtypes = (ctypes.c_int64,)
build_datetime_ns.restype = ctypes.c_void_p

build_datetime_from_utc_ns = lib.build_datetime_from_utc_ns
build_datetime_from_utc_ns.argtypes = (ctypes.c_int64,)
build_datetime_from_utc_ns.restype = ctypes.c_void_p

local_timestamp_nanos = lib.local_timestamp_nanos
local_timestamp_nanos.argtypes = (ctypes.c_void_p,)
local_timestamp_nanos.restype = ctypes.c_int64

timestamp_nanos = lib.timestamp_nanos
timestamp_nanos.argtypes = (ctypes.c_void_p,)
timestamp_nanos.restype = ctypes.c_int64

utc_timestamp_to_local = lib.utc_timestamp_to_local
utc_timestamp_to_local.argtypes = (ctypes.c_int64,)
utc_timestamp_to_local.restype = ctypes.c_int64

_free_datetime = lib.free_datetime
_free_datetime.argtypes = (ctypes.c_void_p,)

get_datetime_year = lib.get_datetime_year
get_datetime_year.argtypes = (ctypes.c_void_p,)
get_datetime_year.restype = ctypes.c_int32

get_datetime_month = lib.get_datetime_month
get_datetime_month.argtypes = (ctypes.c_void_p,)
get_datetime_month.restype = ctypes.c_int32

get_datetime_day = lib.get_datetime_day
get_datetime_day.argtypes = (ctypes.c_void_p,)
get_datetime_day.restype = ctypes.c_int32

get_datetime_hour = lib.get_datetime_hour
get_datetime_hour.argtypes = (ctypes.c_void_p,)
get_datetime_hour.restype = ctypes.c_int32

get_datetime_minute = lib.get_datetime_minute
get_datetime_minute.argtypes = (ctypes.c_void_p,)
get_datetime_minute.restype = ctypes.c_int32

get_datetime_second = lib.get_datetime_second
get_datetime_second.argtypes = (ctypes.c_void_p,)
get_datetime_second.restype = ctypes.c_int32

get_datetime_nanosecond = lib.get_datetime_nanosecond
get_datetime_nanosecond.argtypes = (ctypes.c_void_p,)
get_datetime_nanosecond.restype = ctypes.c_int32

datetime_with_time = lib.datetime_with_time
datetime_with_time.argtypes = (ctypes.c_void_p, (ctypes.c_uint32 * 6))
datetime_with_time.restype = ctypes.c_void_p



import datetime
import operator

from llvmlite import ir
from numba import types
from numba.core import cgutils
from numba.extending import (
    NativeValue,
    as_numba_type,
    box,
    lower_builtin,
    make_attribute_wrapper,
    models,
    overload,
    register_model,
    type_callable,
    typeof_impl,
    unbox,
)

from .ir_utils import long_as_ulong


class DateType(types.Type):
    def __init__(self):
        super().__init__(name="Date")


date_type = DateType()
as_numba_type.register(datetime.date, date_type)


@typeof_impl.register(datetime.date)
def typeof_datetime_date(val, c):
    return date_type


@type_callable(datetime.date)
def type_date(context):
    def typer(year, month, day):
        return date_type

    return typer


@register_model(DateType)
class DateModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("year", types.uint32),
            ("month", types.uint32),
            ("day", types.uint32),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


make_attribute_wrapper(DateType, "year", "year")
make_attribute_wrapper(DateType, "month", "month")
make_attribute_wrapper(DateType, "day", "day")


@lower_builtin(datetime.date, types.uint32, types.uint32, types.uint32)
@lower_builtin(datetime.date, types.int32, types.int32, types.int32)
def datetime_date_constructor_u32(context, builder, sig, args):
    date_type = sig.return_type
    date = cgutils.create_struct_proxy(date_type)(context, builder)
    date.year = args[0]
    date.month = args[1]
    date.day = args[2]
    return date._getvalue()


@lower_builtin(datetime.date, types.int64, types.int64, types.int64)
def datetime_date_constructor_i64(context, builder, sig, args):
    date_type = sig.return_type
    date = cgutils.create_struct_proxy(date_type)(context, builder)
    date.year = builder.trunc(args[0], ir.IntType(32))
    date.month = builder.trunc(args[1], ir.IntType(32))
    date.day = builder.trunc(args[2], ir.IntType(32))
    return date._getvalue()


@unbox(DateType)
def unbox_date(typ, obj, c):
    year_obj = c.pyapi.object_getattr_string(obj, "year")
    year = long_as_ulong(c.pyapi, year_obj)
    month_obj = c.pyapi.object_getattr_string(obj, "month")
    month = long_as_ulong(c.pyapi, month_obj)
    day_obj = c.pyapi.object_getattr_string(obj, "day")
    second = long_as_ulong(c.pyapi, day_obj)
    c.pyapi.decref(year_obj)
    c.pyapi.decref(month_obj)
    c.pyapi.decref(day_obj)

    date = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    date.year = c.builder.trunc(year, ir.IntType(32))
    date.month = c.builder.trunc(month, ir.IntType(32))
    date.day = c.builder.trunc(second, ir.IntType(32))
    # Check for errors
    is_error = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(date._getvalue(), is_error=is_error)


@box(DateType)
def box_date(typ, val, c):
    """
    Box a native date object into a Python datetime.date object.
    """
    date = cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
    year_obj = c.pyapi.long_from_unsigned_int(date.year)
    month_obj = c.pyapi.long_from_unsigned_int(date.month)
    day_obj = c.pyapi.long_from_unsigned_int(date.day)

    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.date))
    date_obj = c.pyapi.call_function_objargs(class_obj, (year_obj, month_obj, day_obj))
    c.pyapi.decref(year_obj)
    c.pyapi.decref(month_obj)
    c.pyapi.decref(day_obj)
    return date_obj


@overload(operator.ge)
def impl_ge(dt, dt2):
    if not isinstance(dt, DateType) or not isinstance(dt, DateType):
        return

    def impl(dt, dt2):
        if dt.year == dt2.year:
            if dt.month == dt2.month:
                return dt.day >= dt2.day
            else:
                return dt.month >= dt2.month
        else:
            return dt.year >= dt2.year

    return impl


@overload(operator.gt)
def impl_gt(dt, dt2):
    if not isinstance(dt, DateType) or not isinstance(dt, DateType):
        return

    def impl(dt, dt2):
        if dt.year == dt2.year:
            if dt.month == dt2.month:
                return dt.day > dt2.day
            else:
                return dt.month > dt2.month
        else:
            return dt.year > dt2.year

    return impl


@overload(operator.le)
def impl_le(dt, dt2):
    if not isinstance(dt, DateType) or not isinstance(dt, DateType):
        return

    def impl(dt, dt2):
        if dt.year == dt2.year:
            if dt.month == dt2.month:
                return dt.day <= dt2.day
            else:
                return dt.month <= dt2.month
        else:
            return dt.year <= dt2.year

    return impl


@overload(operator.lt)
def impl_lt(dt, dt2):
    if not isinstance(dt, DateType) or not isinstance(dt, DateType):
        return

    def impl(dt, dt2):
        if dt.year == dt2.year:
            if dt.month == dt2.month:
                return dt.day < dt2.day
            else:
                return dt.month < dt2.month
        else:
            return dt.year < dt2.year

    return impl


@overload(operator.eq)
def impl_eq(dt, dt2):
    if not isinstance(dt, DateType) or not isinstance(dt, DateType):
        return

    def impl(dt, dt2):
        return dt.year == dt2.year and dt.month == dt2.month and dt.day == dt2.day

    return impl


@overload(operator.ne)
def impl_ne(dt, dt2):
    if not isinstance(dt, DateType) or not isinstance(dt, DateType):
        return

    def impl(dt, dt2):
        return dt.year != dt2.year or dt.month != dt2.month or dt.day != dt2.day

    return impl



from __future__ import annotations

from typing import TYPE_CHECKING

import polars as pl

from .pl import Bonds as PlBonds
from .pl import Futures as PlFutures
from .pl import TfEvaluators as PlTfEvaluators

if TYPE_CHECKING:
    import pandas as pd


class TfEvaluators:
    """
    Treasury Futures Evaluators for pandas DataFrames.

    A pandas-compatible wrapper around the Polars-based TfEvaluators that provides
    various financial calculations for treasury futures and bonds analysis.

    This class converts pandas inputs to Polars for computation and returns
    pandas Series results for seamless integration with pandas workflows.
    """

    def __init__(
        self,
        future: str | pd.Series,
        bond: str | pd.Series,
        date: str | pd.Series,
        future_price: pd.Series | None = None,
        bond_ytm: pd.Series | None = None,
        capital_rate: float | pd.Series | None = None,
        reinvest_rate: float | None = None,
        ctd_bond: str | pd.Series | None = None,
        ctd_ytm: pd.Series | None = None,
    ):
        """
        Initialize TfEvaluators with market data.

        Args:
            future: Future contract code(s)
            bond: Bond code(s)
            date: Evaluation date(s)
            future_price: Future price(s)
            bond_ytm: Bond yield to maturity
            capital_rate: Capital cost rate for carry calculations
            reinvest_rate: Reinvestment rate for coupon payments (optional)
            ctd_bond: CTD bond code(s) (optional, for neutral_net_basis_spread)
            ctd_ytm: CTD bond yield to maturity (optional, for neutral_net_basis_spread)
        """
        self.pl_df = pl.DataFrame(
            {
                "future": future,
                "bond": bond,
                "date": date,
                "future_price": future_price,
                "bond_ytm": bond_ytm,
                "capital_rate": capital_rate,
                "ctd_bond": ctd_bond,
                "ctd_ytm": ctd_ytm,
            }
        )
        self._evaluators = PlTfEvaluators(
            future_price="future_price",
            bond_ytm="bond_ytm",
            capital_rate="capital_rate",
            reinvest_rate=reinvest_rate,
            ctd_bond="ctd_bond",
            ctd_ytm="ctd_ytm",
        )

    @property
    def net_basis_spread(self):
        """
        Calculate net basis spread (净基差).

        Net basis spread = basis spread - carry return

        Returns:
            pd.Series: Net basis spread values
        """
        return self.pl_df.select(net_basis_spread=self._evaluators.net_basis_spread)[
            "net_basis_spread"
        ].to_pandas()

    @property
    def accrued_interest(self):
        """
        Calculate accrued interest (应计利息).

        Returns:
            pd.Series: Accrued interest values
        """
        return self.pl_df.select(accrued_interest=self._evaluators.accrued_interest)[
            "accrued_interest"
        ].to_pandas()

    @property
    def deliver_accrued_interest(self):
        """
        Calculate delivery accrued interest (国债期货交割应计利息).

        Returns:
            pd.Series: Delivery accrued interest values
        """
        return self.pl_df.select(
            deliver_accrued_interest=self._evaluators.deliver_accrued_interest
        )["deliver_accrued_interest"].to_pandas()

    @property
    def cf(self):
        """
        Calculate conversion factor (转换因子).

        Returns:
            pd.Series: Conversion factor values
        """
        return self.pl_df.select(cf=self._evaluators.cf)["cf"].to_pandas()

    @property
    def dirty_price(self):
        """
        Calculate bond dirty price (债券全价).

        Returns:
            pd.Series: Bond dirty price values
        """
        return self.pl_df.select(dirty_price=self._evaluators.dirty_price)[
            "dirty_price"
        ].to_pandas()

    @property
    def clean_price(self):
        """
        Calculate bond clean price (债券净价).

        Returns:
            pd.Series: Bond clean price values
        """
        return self.pl_df.select(clean_price=self._evaluators.clean_price)[
            "clean_price"
        ].to_pandas()

    @property
    def future_dirty_price(self):
        """
        Calculate future dirty price (期货全价/发票价格).

        Returns:
            pd.Series: Future dirty price values
        """
        return self.pl_df.select(
            future_dirty_price=self._evaluators.future_dirty_price
        )["future_dirty_price"].to_pandas()

    @property
    def deliver_cost(self):
        """
        Calculate delivery cost (交割成本).

        Delivery cost = bond dirty price - interim coupon payments

        Returns:
            pd.Series: Delivery cost values
        """
        return self.pl_df.select(deliver_cost=self._evaluators.deliver_cost)[
            "deliver_cost"
        ].to_pandas()

    @property
    def basis_spread(self):
        """
        Calculate basis spread (基差).

        Returns:
            pd.Series: Basis spread values
        """
        return self.pl_df.select(basis_spread=self._evaluators.basis_spread)[
            "basis_spread"
        ].to_pandas()

    @property
    def f_b_spread(self):
        """
        Calculate futures-bond spread (期现价差).

        Returns:
            pd.Series: Futures-bond spread values
        """
        return self.pl_df.select(f_b_spread=self._evaluators.f_b_spread)[
            "f_b_spread"
        ].to_pandas()

    @property
    def carry(self):
        """
        Calculate carry return (持有收益).

        Carry return = (delivery accrued - trading accrued + interim coupons) +
                      capital cost rate * (weighted average interim coupons - bond dirty price * remaining days / 365)

        Returns:
            pd.Series: Carry return values
        """
        return self.pl_df.select(carry=self._evaluators.carry)["carry"].to_pandas()

    @property
    def duration(self):
        """
        Calculate modified duration (修正久期).

        Returns:
            pd.Series: Modified duration values
        """
        return self.pl_df.select(duration=self._evaluators.duration)[
            "duration"
        ].to_pandas()

    @property
    def irr(self):
        """
        Calculate internal rate of return (内部收益率).

        Returns:
            pd.Series: Internal rate of return values
        """
        return self.pl_df.select(irr=self._evaluators.irr)["irr"].to_pandas()

    @property
    def future_ytm(self):
        """
        Calculate futures implied yield to maturity (期货隐含收益率).

        Returns:
            pd.Series: Futures implied yield to maturity values
        """
        return self.pl_df.select(future_ytm=self._evaluators.future_ytm)[
            "future_ytm"
        ].to_pandas()

    @property
    def remain_cp_to_deliver(self):
        """
        Calculate remaining coupon payments to delivery (到交割的期间付息).

        Returns:
            pd.Series: Remaining coupon payments to delivery values
        """
        return self.pl_df.select(
            remain_cp_to_deliver=self._evaluators.remain_cp_to_deliver
        )["remain_cp_to_deliver"].to_pandas()

    @property
    def remain_cp_to_deliver_wm(self):
        """
        Calculate weighted average remaining coupon payments to delivery (加权平均到交割的期间付息).

        Returns:
            pd.Series: Weighted average remaining coupon payments to delivery values
        """
        return self.pl_df.select(
            remain_cp_to_deliver_wm=self._evaluators.remain_cp_to_deliver_wm
        )["remain_cp_to_deliver_wm"].to_pandas()

    @property
    def remain_cp_num(self):
        """
        Calculate remaining number of coupon payments (债券剩余付息次数).

        Returns:
            pd.Series: Remaining number of coupon payments values
        """
        return self.pl_df.select(remain_cp_num=self._evaluators.remain_cp_num)[
            "remain_cp_num"
        ].to_pandas()

    @property
    def deliver_date(self):
        """
        Calculate delivery date (交割日).

        Returns:
            pd.Series: Delivery date values
        """
        return self.pl_df.select(deliver_date=self._evaluators.deliver_date)[
            "deliver_date"
        ].to_pandas()

    @property
    def last_trading_date(self):
        """
        Calculate last trading date (最后交易日).

        Returns:
            pd.Series: Last trading date values
        """
        return self.pl_df.select(last_trading_date=self._evaluators.last_trading_date)[
            "last_trading_date"
        ].to_pandas()

    @property
    def remain_year(self):
        return self.pl_df.select(remain_year=self._evaluators.remain_year)[
            "remain_year"
        ].to_pandas()

    @property
    def neutral_net_basis_spread(self):
        """
        Calculate DV-neutral net basis spread (DV中性净基差).

        DV-neutral net basis spread = P - CF_Neutral * F - Carry

        Returns:
            pd.Series: DV-neutral net basis spread values
        """
        return self.pl_df.select(
            neutral_net_basis_spread=self._evaluators.neutral_net_basis_spread
        )["neutral_net_basis_spread"].to_pandas()


class Bonds:
    """
    Bond calculations for pandas DataFrames.

    A pandas-compatible wrapper around the Polars-based Bonds class that provides
    bond-specific financial calculations without requiring futures contract information.
    """

    def __init__(self, bond: str | pd.Series):
        """
        Initialize Bonds with bond identifier.

        Args:
            bond: Bond code(s)
        """
        self.bond = bond

    def accrued_interest(self, date: str | pd.Series):
        """
        Calculate accrued interest for the bond (应计利息).

        Args:
            date: Evaluation date(s)

        Returns:
            pd.Series: Accrued interest values
        """
        df = pl.DataFrame({"bond": self.bond, "date": date})
        return df.select(accrued_interest=PlBonds("bond").accrued_interest("date"))[
            "accrued_interest"
        ].to_pandas()

    def clean_price(self, ytm: float | pd.Series, date: str | pd.Series):
        """
        Calculate bond clean price (债券净价).

        Args:
            ytm: Yield to maturity
            date: Evaluation date(s)

        Returns:
            pd.Series: Bond clean price values
        """
        df = pl.DataFrame({"bond": self.bond, "ytm": ytm, "date": date})
        return df.select(clean_price=PlBonds("bond").clean_price("ytm", "date"))[
            "clean_price"
        ].to_pandas()

    def dirty_price(self, ytm: float | pd.Series, date: str | pd.Series):
        """
        Calculate bond dirty price (债券全价).

        Args:
            ytm: Yield to maturity
            date: Evaluation date(s)

        Returns:
            pd.Series: Bond dirty price values
        """
        df = pl.DataFrame({"bond": self.bond, "ytm": ytm, "date": date})
        return df.select(dirty_price=PlBonds("bond").dirty_price("ytm", "date"))[
            "dirty_price"
        ].to_pandas()

    def duration(self, ytm: float | pd.Series, date: str | pd.Series):
        """
        Calculate modified duration (修正久期).

        Args:
            ytm: Yield to maturity
            date: Evaluation date(s)

        Returns:
            pd.Series: Modified duration values
        """
        df = pl.DataFrame({"bond": self.bond, "ytm": ytm, "date": date})
        return df.select(duration=PlBonds("bond").duration("ytm", "date"))[
            "duration"
        ].to_pandas()

    def remain_cp_num(self, date: str | pd.Series):
        """
        Calculate remaining number of coupon payments (债券剩余付息次数).

        Args:
            date: Evaluation date(s)

        Returns:
            pd.Series: Remaining number of coupon payments values
        """
        df = pl.DataFrame({"bond": self.bond, "date": date})
        return df.select(remain_cp_num=PlBonds("bond").remain_cp_num("date"))[
            "remain_cp_num"
        ].to_pandas()

    def remain_year(self, date: str | pd.Series):
        df = pl.DataFrame({"bond": self.bond, "date": date})
        return df.select(remain_year=PlBonds("bond").remain_year("date"))[
            "remain_year"
        ].to_pandas()

    def carry_date(self):
        df = pl.DataFrame({"bond": self.bond})
        return df.select(carry_date=PlBonds("bond").carry_date())[
            "carry_date"
        ].to_pandas()

    def maturity_date(self):
        df = pl.DataFrame({"bond": self.bond})
        return df.select(maturity_date=PlBonds("bond").maturity_date())[
            "maturity_date"
        ].to_pandas()

    def calc_ytm_with_bond_price(
        self,
        date: str | pd.Series,
        dirty_price: float | pd.Series | None = None,
        clean_price: float | pd.Series | None = None,
    ):
        if clean_price is not None:
            assert dirty_price is None, (
                "should not set dirty_price when clean_price is set"
            )
            price = clean_price
            kwargs = {"clean_price": "price"}
        else:
            assert clean_price is None, (
                "should not set clean_price when dirty_price is set"
            )
            assert dirty_price is not None
            price = dirty_price
            kwargs = {"dirty_price": "price"}
        df = pl.DataFrame({"bond": self.bond, "price": price, "date": date})
        return df.select(
            ytm=PlBonds("bond").calc_ytm_with_price(date="date", **kwargs)
        )["ytm"].to_pandas()


class Futures:
    def __init__(self, future: str | pd.Series):
        self.future = future

    def deliver_date(self):
        """
        Calculate delivery date (交割日).

        Args:
            date: Evaluation date(s)

        Returns:
            pd.Series: Delivery date values
        """
        df = pl.DataFrame({"future": self.future})
        return df.select(deliver_date=PlFutures("future").deliver_date())[
            "deliver_date"
        ].to_pandas()

    def last_trading_date(self):
        """
        Calculate last trading date (最后交易日).

        Args:
            date: Evaluation date(s)

        Returns:
            pd.Series: Last trading date values
        """
        df = pl.DataFrame({"future": self.future})
        return df.select(last_trading_date=PlFutures("future").last_trading_date())[
            "last_trading_date"
        ].to_pandas()


def find_workday(date: str | pd.Series, market: str, offset: int = 0):
    """
    Find the workday based on the given date and market calendar.

    Args:
        date: Input date(s)
        market: Market identifier ("IB" or "SSE")
        offset: Number of workdays to offset (default: 0)

    Returns:
        pd.Series: Adjusted workday values
    """
    from .pl import find_workday as pl_find_workday

    df = pl.DataFrame({"date": date}).select(pl.col("date").dt.date())
    return df.select(workday=pl_find_workday("date", market, offset))[
        "workday"
    ].to_pandas()


def is_business_day(date: str | pd.Series, market: str):
    """
    Check if the given date is a business day for the specified market.

    Args:
        date: Input date(s)
        market: Market identifier ("IB" or "SSE")

    Returns:
        pd.Series: Boolean values indicating if dates are business days
    """
    from .pl import is_business_day as pl_is_business_day

    df = pl.DataFrame({"date": date}).select(pl.col("date").dt.date())
    return df.select(is_business=pl_is_business_day("date", market))[
        "is_business"
    ].to_pandas()



from llvmlite import ir
from numba import types
from numba.core import cgutils, utils
from numba.extending import (
    as_numba_type,
    box,
    intrinsic,
    lower_builtin,
    make_attribute_wrapper,
    models,
    overload_attribute,
    overload_method,
    register_model,
    type_callable,
    typeof_impl,
)

from pybond import TfEvaluator
from pybond.ffi import (
    tf_evaluator_accrued_interest,
    tf_evaluator_basis_spread,
    tf_evaluator_bond_ytm,
    tf_evaluator_calc_all,
    tf_evaluator_capital_rate,
    tf_evaluator_carry,
    tf_evaluator_cf,
    tf_evaluator_clean_price,
    tf_evaluator_deliver_accrued_interest,
    tf_evaluator_deliver_cost,
    tf_evaluator_dirty_price,
    tf_evaluator_duration,
    tf_evaluator_f_b_spread,
    tf_evaluator_future_dirty_price,
    tf_evaluator_future_price,
    tf_evaluator_future_ytm,
    tf_evaluator_irr,
    tf_evaluator_is_deliverable,
    tf_evaluator_net_basis_spread,
    tf_evaluator_reinvest_rate,
    tf_evaluator_remain_cp_num,
    tf_evaluator_remain_cp_to_deliver,
    tf_evaluator_remain_cp_to_deliver_wm,
    tf_evaluator_remain_days_to_deliver,
)

from .nb_date import date_type


class TfEvaluatorType(types.Type):
    def __init__(self):
        super().__init__(name="TfEvaluator")


tf_evaluator_type = TfEvaluatorType()
as_numba_type.register(TfEvaluator, tf_evaluator_type)


@typeof_impl.register(TfEvaluator)
def typeof_tf_evaluator(val, c):
    return tf_evaluator_type


@type_callable(TfEvaluator)
def type_tf_evaluator(context):
    def typer(
        future,
        bond,
        date=None,
        future_price=float("nan"),
        bond_ytm=float("nan"),
        capital_rate=float("nan"),
        reinvest_rate=None,
    ):
        return tf_evaluator_type

    return typer


@register_model(TfEvaluatorType)
class TfEvaluatorModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("ptr", types.voidptr),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


make_attribute_wrapper(TfEvaluatorType, "ptr", "ptr")


@lower_builtin(
    TfEvaluator,
    types.string,
    types.string,
    date_type,
    types.float64,
    types.float64,
    types.float64,
)
def impl_tf_evaluator_builder(context, builder, sig, args):
    typ = sig.return_type
    (
        future,
        bond,
        date_val,
        future_price,
        bond_ytm,
        capital_rate,
    ) = args

    # Get string data from Numba strings
    future_code = context.make_helper(builder, types.string, future)
    bond_code = context.make_helper(builder, types.string, bond)
    date = context.make_helper(builder, date_type, date_val)

    # Call create_tf_evaluator
    fn = cgutils.get_or_insert_function(
        builder.module,
        ir.FunctionType(
            ir.PointerType(ir.IntType(8)),
            [
                ir.PointerType(ir.IntType(8)),  # future_code_ptr
                ir.IntType(utils.MACHINE_BITS),  # future_code_len
                ir.DoubleType(),  # future_price
                ir.PointerType(ir.IntType(8)),  # bond_code_ptr
                ir.IntType(utils.MACHINE_BITS),  # bond_code_len
                ir.DoubleType(),  # bond_ytm
                ir.DoubleType(),  # capital_rate
                ir.IntType(32),  # year
                ir.IntType(32),  # month
                ir.IntType(32),  # day
            ],
        ),
        name="create_tf_evaluator",
    )

    ptr = builder.call(
        fn,
        [
            future_code.data,
            future_code.length,
            future_price,
            bond_code.data,
            bond_code.length,
            bond_ytm,
            capital_rate,
            date.year,
            date.month,
            date.day,
        ],
    )

    # Create TfEvaluator object
    evaluator = cgutils.create_struct_proxy(typ)(context, builder)
    evaluator.ptr = ptr
    return evaluator._getvalue()


@lower_builtin(
    TfEvaluator,
    types.string,
    types.string,
    date_type,
    types.float64,
    types.float64,
    types.float64,
    types.float64,
)
def impl_tf_evaluator_builder_with_reinvest(context, builder, sig, args):
    typ = sig.return_type
    (
        future_code_val,
        bond_code_val,
        date_val,
        future_price_val,
        bond_ytm_val,
        capital_rate_val,
        reinvest_rate_val,
    ) = args

    # Get string data from Numba strings
    future_code = context.make_helper(builder, types.string, future_code_val)
    bond_code = context.make_helper(builder, types.string, bond_code_val)

    # Get date components
    date = context.make_helper(builder, date_type, date_val)

    # Call create_tf_evaluator_with_reinvest
    fn = cgutils.get_or_insert_function(
        builder.module,
        ir.FunctionType(
            ir.PointerType(ir.IntType(8)),
            [
                ir.PointerType(ir.IntType(8)),  # future_code_ptr
                ir.IntType(utils.MACHINE_BITS),  # future_code_len
                ir.DoubleType(),  # future_price
                ir.PointerType(ir.IntType(8)),  # bond_code_ptr
                ir.IntType(utils.MACHINE_BITS),  # bond_code_len
                ir.DoubleType(),  # bond_ytm
                ir.DoubleType(),  # capital_rate
                ir.DoubleType(),  # reinvest_rate
                ir.IntType(32),  # year
                ir.IntType(32),  # month
                ir.IntType(32),  # day
            ],
        ),
        name="create_tf_evaluator_with_reinvest",
    )

    ptr = builder.call(
        fn,
        [
            future_code.data,
            future_code.length,
            future_price_val,
            bond_code.data,
            bond_code.length,
            bond_ytm_val,
            capital_rate_val,
            reinvest_rate_val,
            date.year,
            date.month,
            date.day,
        ],
    )

    # Create TfEvaluator object
    evaluator = cgutils.create_struct_proxy(typ)(context, builder)
    evaluator.ptr = ptr
    return evaluator._getvalue()


# Property getters
@overload_attribute(TfEvaluatorType, "is_deliverable")
def tf_evaluator_attr_is_deliverable(evaluator):
    def impl(evaluator):
        return tf_evaluator_is_deliverable(evaluator.ptr) == 1

    return impl


@overload_attribute(TfEvaluatorType, "bond_ytm")
def tf_evaluator_attr_bond_ytm(evaluator):
    def impl(evaluator):
        return tf_evaluator_bond_ytm(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "future_price")
def tf_evaluator_attr_future_price(evaluator):
    def impl(evaluator):
        return tf_evaluator_future_price(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "capital_rate")
def tf_evaluator_attr_capital_rate(evaluator):
    def impl(evaluator):
        return tf_evaluator_capital_rate(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "reinvest_rate")
def tf_evaluator_attr_reinvest_rate(evaluator):
    def impl(evaluator):
        return tf_evaluator_reinvest_rate(evaluator.ptr)

    return impl


# Calculation methods
@overload_attribute(TfEvaluatorType, "accrued_interest")
def tf_evaluator_method_accrued_interest(evaluator):
    def impl(evaluator):
        return tf_evaluator_accrued_interest(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "deliver_accrued_interest")
def tf_evaluator_method_deliver_accrued_interest(evaluator):
    def impl(evaluator):
        return tf_evaluator_deliver_accrued_interest(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "cf")
def tf_evaluator_method_cf(evaluator):
    def impl(evaluator):
        return tf_evaluator_cf(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "dirty_price")
def tf_evaluator_method_dirty_price(evaluator):
    def impl(evaluator):
        return tf_evaluator_dirty_price(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "clean_price")
def tf_evaluator_method_clean_price(evaluator):
    def impl(evaluator):
        return tf_evaluator_clean_price(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "future_dirty_price")
def tf_evaluator_method_future_dirty_price(evaluator):
    def impl(evaluator):
        return tf_evaluator_future_dirty_price(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "deliver_cost")
def tf_evaluator_method_deliver_cost(evaluator):
    def impl(evaluator):
        return tf_evaluator_deliver_cost(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "basis_spread")
def tf_evaluator_method_basis_spread(evaluator):
    def impl(evaluator):
        return tf_evaluator_basis_spread(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "f_b_spread")
def tf_evaluator_method_f_b_spread(evaluator):
    def impl(evaluator):
        return tf_evaluator_f_b_spread(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "carry")
def tf_evaluator_method_carry(evaluator):
    def impl(evaluator):
        return tf_evaluator_carry(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "net_basis_spread")
def tf_evaluator_method_net_basis_spread(evaluator):
    def impl(evaluator):
        return tf_evaluator_net_basis_spread(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "duration")
def tf_evaluator_method_duration(evaluator):
    def impl(evaluator):
        return tf_evaluator_duration(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "irr")
def tf_evaluator_method_irr(evaluator):
    def impl(evaluator):
        return tf_evaluator_irr(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "future_ytm")
def tf_evaluator_method_future_ytm(evaluator):
    def impl(evaluator):
        return tf_evaluator_future_ytm(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "remain_days_to_deliver")
def tf_evaluator_method_remain_days_to_deliver(evaluator):
    def impl(evaluator):
        return tf_evaluator_remain_days_to_deliver(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "remain_cp_num")
def tf_evaluator_method_remain_cp_num(evaluator):
    def impl(evaluator):
        return tf_evaluator_remain_cp_num(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "remain_cp_to_deliver")
def tf_evaluator_method_remain_cp_to_deliver(evaluator):
    def impl(evaluator):
        return tf_evaluator_remain_cp_to_deliver(evaluator.ptr)

    return impl


@overload_attribute(TfEvaluatorType, "remain_cp_to_deliver_wm")
def tf_evaluator_method_remain_cp_to_deliver_wm(evaluator):
    def impl(evaluator):
        return tf_evaluator_remain_cp_to_deliver_wm(evaluator.ptr)

    return impl


@overload_method(TfEvaluatorType, "calc_all")
def tf_evaluator_method_calc_all(evaluator):
    def impl(evaluator):
        _ = tf_evaluator_calc_all(evaluator.ptr)
        return evaluator

    return impl


@box(TfEvaluatorType)
def box_tf_evaluator(typ, val, c):
    """
    Convert a native TfEvaluator structure to python TfEvaluator.
    """
    evaluator = cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)

    # Convert the pointer to usize for from_ptr method
    ptr_as_usize = c.builder.ptrtoint(evaluator.ptr, ir.IntType(64))
    ptr_obj = c.pyapi.from_native_value(types.uint64, ptr_as_usize)

    # Get TfEvaluator class and call from_ptr
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(TfEvaluator))
    from_ptr_method = c.pyapi.object_getattr_string(class_obj, "from_ptr")

    # Call TfEvaluator.from_ptr(ptr)
    res = c.pyapi.call_function_objargs(from_ptr_method, (ptr_obj,))

    # Clean up Python objects
    c.pyapi.decref(ptr_obj)
    c.pyapi.decref(class_obj)
    c.pyapi.decref(from_ptr_method)

    return res


@intrinsic
def _tf_evaluator_update_call(
    typingctx,
    evaluator_t,
    future_price_t,
    bond_ytm_t,
    date_t,
    future_code_t,
    bond_code_t,
    capital_rate_t,
):
    """Intrinsic for calling tf_evaluator_update_info FFI function."""

    def codegen(context, builder, sig, args):
        (
            evaluator_val,
            future_price_val,
            bond_ytm_val,
            date_val,
            future_code_val,
            bond_code_val,
            capital_rate_val,
        ) = args

        # Get existing evaluator
        evaluator = cgutils.create_struct_proxy(tf_evaluator_type)(
            context, builder, value=evaluator_val
        )

        # Get string data from Numba strings
        future_code = context.make_helper(builder, types.string, future_code_val)
        bond_code = context.make_helper(builder, types.string, bond_code_val)

        # Get date components
        date = context.make_helper(builder, date_type, date_val)

        # Call tf_evaluator_update_info
        fn = cgutils.get_or_insert_function(
            builder.module,
            ir.FunctionType(
                ir.IntType(32),  # return type: i32
                [
                    ir.PointerType(ir.IntType(8)),  # evaluator ptr
                    ir.PointerType(ir.IntType(8)),  # future_code_ptr
                    ir.IntType(utils.MACHINE_BITS),  # future_code_len
                    ir.DoubleType(),  # future_price
                    ir.PointerType(ir.IntType(8)),  # bond_code_ptr
                    ir.IntType(utils.MACHINE_BITS),  # bond_code_len
                    ir.DoubleType(),  # bond_ytm
                    ir.DoubleType(),  # capital_rate
                    ir.IntType(32),  # year
                    ir.IntType(32),  # month
                    ir.IntType(32),  # day
                ],
            ),
            name="tf_evaluator_update_info",
        )

        result = builder.call(
            fn,
            [
                evaluator.ptr,
                future_code.data,
                future_code.length,
                future_price_val,
                bond_code.data,
                bond_code.length,
                bond_ytm_val,
                capital_rate_val,
                date.year,
                date.month,
                date.day,
            ],
        )

        return result

    sig = types.int32(
        evaluator_t,
        future_price_t,
        bond_ytm_t,
        date_t,
        future_code_t,
        bond_code_t,
        capital_rate_t,
    )
    return sig, codegen


@overload_method(TfEvaluatorType, "update")
# 采用和update pyi同样的参数顺序
def tf_evaluator_method_update(
    evaluator, future_price, bond_ytm, date, future, bond, capital_rate
):
    def update_impl(
        evaluator, future_price, bond_ytm, date, future, bond, capital_rate
    ):
        _result = _tf_evaluator_update_call(
            evaluator, future_price, bond_ytm, date, future, bond, capital_rate
        )
        # Return the evaluator itself (it's been updated in-place)
        return evaluator

    return update_impl



import ctypes

from .lib import lib

create_bond = lib.create_bond
create_bond.argtypes = (ctypes.c_void_p, ctypes.c_size_t)
create_bond.restype = ctypes.c_void_p

free_bond = lib.free_bond
free_bond.argtypes = [ctypes.c_void_p]
free_bond.restype = None

bond_coupon_rate = lib.bond_coupon_rate
bond_coupon_rate.argtypes = [ctypes.c_void_p]
bond_coupon_rate.restype = ctypes.c_double

bond_full_code = lib.bond_full_code
bond_full_code.argtypes = [ctypes.c_void_p]
bond_full_code.restype = ctypes.c_char_p

bond_calc_ytm = lib.bond_calc_ytm
bond_calc_ytm.argtypes = [
    ctypes.c_void_p,
    ctypes.c_double,
    ctypes.c_uint32,
    ctypes.c_uint32,
    ctypes.c_uint32,
]
bond_calc_ytm.restype = ctypes.c_double

bond_duration = lib.bond_duration
bond_duration.argtypes = [
    ctypes.c_void_p,
    ctypes.c_double,
    ctypes.c_uint32,
    ctypes.c_uint32,
    ctypes.c_uint32,
]
bond_duration.restype = ctypes.c_double

bond_accrued_interest = lib.bond_accrued_interest
bond_accrued_interest.argtypes = [
    ctypes.c_void_p,
    ctypes.c_uint32,
    ctypes.c_uint32,
    ctypes.c_uint32,
]
bond_accrued_interest.restype = ctypes.c_double

bond_dirty_price = lib.bond_dirty_price
bond_dirty_price.argtypes = [
    ctypes.c_void_p,
    ctypes.c_double,
    ctypes.c_uint32,
    ctypes.c_uint32,
    ctypes.c_uint32,
]
bond_dirty_price.restype = ctypes.c_double

bond_clean_price = lib.bond_clean_price
bond_clean_price.argtypes = [
    ctypes.c_void_p,
    ctypes.c_double,
    ctypes.c_uint32,
    ctypes.c_uint32,
    ctypes.c_uint32,
]
bond_clean_price.restype = ctypes.c_double



from __future__ import annotations

import os
from importlib.util import find_spec
from pathlib import Path
from typing import TYPE_CHECKING

from .pybond import Bond as _BondRS
from .pybond import Future, download_bond

if TYPE_CHECKING:
    from datetime import date


WIND_AVAILABLE = find_spec("WindPy") is not None

if os.environ.get("BONDS_INFO_PATH") is not None:
    bonds_info_environ_flag = True
    bonds_info_path = Path(os.environ.get("BONDS_INFO_PATH"))
else:
    bonds_info_environ_flag = False
    old_default_path = Path(__file__).parent / "data" / "bonds_info"
    bonds_info_path = Path.home() / "tea-bond" / "bonds_info"
    if old_default_path.exists() and not bonds_info_path.exists():
        import shutil

        shutil.move(str(old_default_path), str(bonds_info_path))
    if not bonds_info_path.exists():
        bonds_info_path.mkdir(parents=True)
    os.environ["BONDS_INFO_PATH"] = str(bonds_info_path)

if not bonds_info_path.exists():
    bonds_info_path.mkdir(parents=True)


class Bond(_BondRS):
    def __new__(
        cls,
        code: str | int = "",
        path: str | Path | None = None,
        *,
        download: bool = True,
    ):
        """
        Create a new Bond instance.

        Args:
            code (str | int): The bond code. If no extension is provided, '.IB' will be appended.
            path (str | Path | None): Path to the bond info file. If None, uses default bonds_info_path.
            download (bool): Whether to automatically download the bond info if it doesn't exist.

        Returns:
            Bond: A new Bond instance, either loaded from existing JSON file or downloaded.

        Note:
            If a JSON file for the bond code doesn't exist at the specified path,
            the bond info will be downloaded automatically.
        """
        code = str(code)
        if code == "":
            return super().__new__(cls, "", path)
        if "." not in code:
            code = code + ".IB"
        try:
            return super().__new__(cls, code, path)
        except ValueError as e:
            if download:
                path = bonds_info_path if path is None else Path(path)
                cls.download(code, path)
                return super().__new__(cls, code, path)
            else:
                raise ValueError from e

    @classmethod
    def from_json(cls, data: str | dict) -> Bond:
        if isinstance(data, str):
            import json

            data = json.loads(data)
        bond = Bond()
        for k, v in data.items():
            setattr(bond, k, v)
        return bond

    @staticmethod
    def download(
        code: str, path: str | None = None, source: str | None = None, save=True
    ):
        """
        Download bond information from a specified source.

        This method downloads bond information for a given bond code from either Wind or Rust.
        If no source is specified, it defaults to Wind if the WindPy module is available; otherwise,
        it falls back to Rust.

        If the source is 'rust', the method will download IB bond information from China Money and
        SH bond information from SSE (Shanghai Stock Exchange).

        Args:
            code (str): The bond code in the format 'XXXXXX.YY'. The code must include a dot.
            path (str | None): The directory path where the downloaded bond information should be saved.
                              If None, the default path is used.
            source (str | None): The source from which to download the bond information. Valid options are
                                'wind' or 'rust'. If None, the source is automatically determined.
            save (bool): Whether to save the downloaded bond information to the specified path.
                        Defaults to True.

        Returns:
            Bond: The downloaded bond object if the source is 'rust' and save is False.
                  Otherwise, returns None.

        Raises:
            AssertionError: If the code is not in the correct format or if the source is invalid.
        """
        if source is None:
            # 优先从wind下载
            source = "wind" if WIND_AVAILABLE else "rust"
        assert "." in code, "code should be in the format of XXXXXX.YY"
        assert source in ("wind", "rust")
        if source == "wind":
            from .download import fetch_symbols, login

            print(f"Start downloading bond info for {code} from Wind")
            login()
            fetch_symbols([code], save=save, save_folder=path)
        else:
            # let rust side handle the download
            print(f"download {code}")
            bond = download_bond(code)
            if save:
                bond.save(path)
            return bond

    def accrued_interest(
        self, date: date, cp_dates: tuple[date, date] | None = None
    ) -> float:
        """
        计算应计利息

        银行间和交易所的计算规则不同,银行间是算头不算尾,而交易所是算头又算尾
        """
        return self.calc_accrued_interest(date, cp_dates=cp_dates)

    def dirty_price(
        self,
        ytm: float,
        date: date,
        cp_dates: tuple[date, date] | None = None,
        remain_cp_num: int | None = None,
    ) -> float:
        """通过ytm计算债券全价"""
        return self.calc_dirty_price_with_ytm(
            ytm, date, cp_dates=cp_dates, remain_cp_num=remain_cp_num
        )

    def clean_price(
        self,
        ytm: float,
        date: date,
        cp_dates: tuple[date, date] | None = None,
        remain_cp_num: int | None = None,
    ) -> float:
        """通过ytm计算债券净价"""
        return self.calc_clean_price_with_ytm(
            ytm, date, cp_dates=cp_dates, remain_cp_num=remain_cp_num
        )

    def macaulay_duration(
        self,
        ytm: float,
        date: date,
        cp_dates: tuple[date, date] | None = None,
        remain_cp_num: int | None = None,
    ) -> float:
        """计算麦考利久期"""
        return self.calc_macaulay_duration(
            ytm, date, cp_dates=cp_dates, remain_cp_num=remain_cp_num
        )

    def duration(
        self,
        ytm: float,
        date: date,
        cp_dates: tuple[date, date] | None = None,
        remain_cp_num: int | None = None,
    ) -> float:
        """计算修正久期"""
        return self.calc_duration(
            ytm, date, cp_dates=cp_dates, remain_cp_num=remain_cp_num
        )

    def cf(self, future: str | Future) -> float:
        """计算转换因子"""
        from .pybond import TfEvaluator

        return TfEvaluator(future, self).cf

    def calc_ytm_with_clean_price(
        self,
        clean_price: float,
        date: date,
        cp_dates: tuple[date, date] | None = None,
        remain_cp_num: int | None = None,
    ) -> float:
        """通过净价计算债券ytm"""
        dirty_price = clean_price + self.accrued_interest(date, cp_dates=cp_dates)
        return self.calc_ytm_with_price(
            dirty_price, date, cp_dates=cp_dates, remain_cp_num=remain_cp_num
        )



from __future__ import annotations

import json
from datetime import date
from decimal import Decimal
from pathlib import Path

from WindPy import w

default_save_folder = Path("bonds_info")


def save_json(path: Path | str, data: dict) -> None:
    """
    Save data into json file in temp path.
    """
    path = Path(path)
    with path.open(mode="w+", encoding="UTF-8") as f:
        json.dump(data, f, indent=4, ensure_ascii=False)


def get_interest_type(typ: str):
    if typ == "固定利率":
        return "Fixed"
    elif typ == "浮动利率":
        return "Floating"
    elif typ == "累进利率":
        return "Progressive"
    elif typ == "零息":
        return "Zero"
    else:
        msg = f"Unknown interest type: {typ}"
        raise ValueError(msg)


def get_payment_type(typ: str):
    if typ == "附息":
        return "Coupon_Bear"
    elif typ == "到期一次还本付息":
        return "One_Time"
    elif typ == "贴现":
        return "Zero_Coupon"
    else:
        msg = f"Unknown payment type: {typ}"
        raise ValueError(msg)


def fetch_symbols(
    symbols: list[str],
    *,
    save: bool = True,
    skip: bool = True,
    save_folder: Path | str | None = None,
):
    if save_folder is None:
        save_folder = default_save_folder
    if isinstance(save_folder, str):
        save_folder = Path(save_folder)
    if skip:
        symbols = [s for s in symbols if not (save_folder / f"{s}.json").exists()]
    data = w.wss(
        symbols,
        "sec_name,carrydate,maturitydate,interesttype,couponrate,paymenttype,actualbenchmark,coupon,interestfrequency,latestpar",
        f"tradeDate={date.today()}",
    ).Data
    returns = []
    for i, symbol in enumerate(symbols):
        m = {"bond_code": symbol}
        m["mkt"] = symbol.split(".")[1].upper()
        m["abbr"] = data[0][i]  # 债券简称
        m["par_value"] = float(data[9][i])  # 面值
        m["cp_type"] = get_payment_type(data[7][i])  # 付息频率
        m["interest_type"] = get_interest_type(data[3][i])  # 付息方式
        m["cp_rate_1st"] = float(Decimal(str(data[4][i])) / 100)  # 票面利率
        m["base_rate"] = None
        m["rate_spread"] = None
        if m["cp_type"] == "Coupon_Bear":
            m["inst_freq"] = int(data[8][i])  # 年付息次数
        elif m["cp_type"] == "One_Time":
            m["inst_freq"] = 1
        elif m["cp_type"] == "Zero_Coupon":
            m["inst_freq"] = 0
        m["carry_date"] = data[1][i].strftime("%Y-%m-%d")  # 起息日
        m["maturity_date"] = data[2][i].strftime("%Y-%m-%d")  # 到期日
        m["day_count"] = data[6][i]  # 实际基准
        returns.append(m)
        print(m)
        if save:
            if not save_folder.exists():
                save_folder.mkdir(parents=True)
            path = save_folder / f"{symbol}.json"
            save_json(path, m)
        return returns


WAIT_LOGIN = False


def login():
    global WAIT_LOGIN
    if w.isconnected():
        return
    if WAIT_LOGIN:
        import time

        time.sleep(0.2)
        login()
    WAIT_LOGIN = True
    login_res = w.start(waitTime=8)
    WAIT_LOGIN = False
    if login_res.ErrorCode != 0:
        msg = f"Failed to login to Wind: {login_res.ErrorCode}"
        raise RuntimeError(msg)


def get_all_symbols():
    sector_ids = (
        # "a101010101000000",  # 国债银行间
        # "a101010104000000",  # 政策性银行债
        "a101010201000000",  # 上交所国债
    )
    res = []
    names = []
    for sector_id in sector_ids:
        all_symbols = w.wset(
            "sectorconstituent", f"sectorid={sector_id};field=wind_code,sec_name"
        ).Data
        res.extend(all_symbols[0])
        names.extend(all_symbols[1])
    print("共有", len(res), "只债券")
    return res


if __name__ == "__main__":
    login()
    # symbols = ["220003.IB", "220021.IB", "220006.IB", "220010.IB"]
    # symbols = ["240006.IB"]

    # symbols = ["019733.SH"]
    # symbols = ["020647.SH"]
    # symbols = ["019727.SH"]
    symbols = ["2400006.IB"]
    # symbols = get_all_symbols()

    fetch_symbols(symbols, save=0, skip=True)



from llvmlite import ir
from numba.core import cgutils


def ir_isinstance(pyapi, obj, typ):
    fnty = ir.FunctionType(ir.IntType(1), [pyapi.pyobj, pyapi.pyobj])
    fn = cgutils.get_or_insert_function(
        pyapi.builder.module, fnty, name="PyObject_IsInstance"
    )
    return pyapi.builder.call(fn, (obj, typ))


def ir_build_datetime(val, builder, *, from_utc: bool = False):
    mod = builder.module
    fnty = ir.FunctionType(ir.PointerType(ir.IntType(8)), [ir.IntType(64)])
    if not from_utc:
        build_datetime_ns_fn = cgutils.get_or_insert_function(
            mod, fnty, "build_datetime_ns"
        )
    else:
        build_datetime_ns_fn = cgutils.get_or_insert_function(
            mod, fnty, "build_datetime_from_utc_ns"
        )
    ptr = builder.call(build_datetime_ns_fn, [val])
    return ptr


def ir_timestamp_nanos(ptr, builder):
    mod = builder.module
    fnty = ir.FunctionType(ir.IntType(64), [ir.PointerType(ir.IntType(8))])
    fn = cgutils.get_or_insert_function(mod, fnty, "timestamp_nanos")
    ptr = builder.call(fn, [ptr])
    return ptr


def ir_local_timestamp_nanos(ptr, builder):
    mod = builder.module
    fnty = ir.FunctionType(ir.IntType(64), [ir.PointerType(ir.IntType(8))])
    fn = cgutils.get_or_insert_function(mod, fnty, "local_timestamp_nanos")
    ptr = builder.call(fn, [ptr])
    return ptr


def long_as_ulong(pyapi, numobj):
    fnty = ir.FunctionType(pyapi.ulong, [pyapi.pyobj])
    fn = pyapi._get_function(fnty, name="PyLong_AsUnsignedLong")
    return pyapi.builder.call(fn, [numobj])



from datetime import date
from pathlib import Path

class Bond:
    """A class representing a bond instrument."""

    def __init__(self, code: str | int, path: str | Path | None = None) -> None:
        """
        Create a new Bond instance.

        Args:
            code: Bond code
            path: Path to directory containing bond data. Defaults to None

        Raises:
            ValueError: If bond data cannot be read or parsed
        """

    def copy(self) -> Bond: ...
    @property
    def code(self) -> str:
        """债券代码, 不包含交易所后缀"""

    @property
    def full_code(self) -> str:
        """债券代码, 包含交易所后缀"""

    @property
    def market(self) -> str:
        """债券市场"""

    @property
    def abbr(self) -> str:
        """债券简称"""

    @property
    def name(self) -> str:
        """债券名称 (alias for abbr)"""

    @property
    def par_value(self) -> float:
        """债券面值"""

    @property
    def coupon_type(self) -> str:
        """息票品种"""

    @property
    def interest_type(self) -> str:
        """息票利率类型"""

    @property
    def coupon_rate(self) -> float:
        """票面利率, 浮动付息债券仅表示发行时票面利率"""

    @property
    def cp_rate(self) -> float:
        """票面利率, 浮动付息债券仅表示发行时票面利率"""

    @property
    def base_rate(self) -> float | None:
        """基准利率, 浮动付息债券适用"""

    @property
    def rate_spread(self) -> float | None:
        """固定利差, 浮动付息债券适用"""

    @property
    def inst_freq(self) -> int:
        """年付息次数"""

    @property
    def carry_date(self) -> date:
        """起息日"""

    @property
    def maturity_date(self) -> date:
        """到期日"""

    @property
    def day_count(self) -> str:
        """计息基准"""

    def is_zero_coupon(self) -> bool:
        """是否为零息债券"""

    def remain_year(self, date: date | None = None) -> float:
        """剩余年数"""

    def issue_year(self) -> int:
        """发行年数"""

    @property
    def get_coupon(self) -> float:
        """
        获取区间付息(单个付息周期的利息金额)

        区间付息 = 票面利率 * 面值 / 年付息次数
        """

    @property
    def last_cp_year_days(self) -> int:
        """最后一个计息年度的天数"""

    def nearest_cp_date(self, date: date) -> tuple[date, date]:
        """获取上一付息日和下一付息日"""

    def remain_cp_num(self, date: date, next_cp_date: date | None = None) -> int:
        """剩余的付息次数"""

    def remain_cp_num_until(
        self, date: date, until_date: date, next_cp_date: date | None = None
    ) -> int:
        """剩余的付息次数直到指定日期"""

    def remain_cp_dates_until(
        self, date: date, until_date: date, next_cp_date: date | None = None
    ) -> list[date]:
        """剩余的付息日期列表"""

    def calc_ytm_with_price(
        self,
        dirty_price: float,
        date: date,
        cp_dates: tuple[date, date] | None = None,
        remain_cp_num: int | None = None,
    ) -> float:
        """通过债券全价计算ytm"""

class Future:
    """A class representing a treasury futures contract."""

    def __init__(self, code: str) -> None:
        """
        Create a new Future instance.

        Args:
            code: Future contract code
        """

    def copy(self) -> Future: ...
    def is_deliverable(
        self,
        carry_date: date,
        maturity_date: date,
        delivery_date: date | None = None,
    ) -> bool:
        """
        判断是否是可交割券

        Args:
            carry_date: 起息日
            maturity_date: 到期日
            delivery_date: 可以传入已计算过的期货配对缴款日避免重复计算
        """

    def last_trading_date(self) -> date:
        """
        计算期货合约的最后交易日

        计算国债期货的最后交易日=合约到期月份的第二个星期五
        根据合约代码, 依据中金所的国债期货合约最后交易日的说, 返回该合约的最后交易日
        获取年月部分
        """

    def deliver_date(self) -> date:
        """
        获取期货合约的配对缴款日

        交割日为3天,其中第2天为缴款日,即最后交易日的第2个交易日,最后交易日一定为周五,所以缴款日一定是一个周二
        """

    def first_trading_date(self) -> date:
        """
        获取期货合约的首个交易日

        对于首批上市合约,返回该品种的上市日期;
        对于后续合约,返回前3季度合约最后交易日的下一个交易日
        """

    def trading_window(self) -> tuple[date, date]:
        """
        获取期货合约的交易区间

        返回 (首个交易日, 最后交易日)
        """

    def code(self) -> str:
        """期货代码"""

    def market(self) -> str | None:
        """期货市场"""

    def next_future(self) -> Future:
        """获取下一季月合约"""

    def prev_future(self) -> Future:
        """获取上一季月合约"""

    @staticmethod
    def trading_futures(
        start: date,
        end: date | None = None,
        future_type: str | None = None,
    ) -> list[str]:
        """获取指定时间段内有交易的期货合约列表"""

    def future_type(self) -> str:
        """获取期货合约的类型"""

class TfEvaluator:
    """A class for evaluating treasury futures."""

    def __init__(
        self,
        future: Future | str,
        bond: Bond | str | int,
        date: date | str | None = None,
        future_price: float = float("nan"),
        bond_ytm: float = float("nan"),
        capital_rate: float = float("nan"),
        reinvest_rate: float | None = None,
    ) -> None: ...
    def copy(self) -> TfEvaluator: ...
    def date(self) -> date:
        """获取计算日期"""

    @property
    def bond_code(self) -> str:
        """获取债券代码"""

    @property
    def bond_ytm(self) -> float:
        """获取债券收益率"""

    @property
    def future(self) -> str:
        """获取期货代码"""

    @property
    def future_price(self) -> float:
        """获取期货价格"""

    @property
    def deliverable(self) -> bool:
        """判断债券是否是期货的可交割券"""

    def with_deliver_date(self) -> TfEvaluator:
        """计算期货配对缴款日"""

    def with_nearest_cp_dates(self) -> TfEvaluator:
        """计算前一付息日和下一付息日"""

    def with_deliver_cp_dates(self) -> TfEvaluator:
        """计算交割日的前一付息日和下一付息日"""

    def with_remain_days_to_deliver(self) -> TfEvaluator:
        """计算到交割的剩余天数"""

    def with_remain_cp_num(self) -> TfEvaluator:
        """计算剩余付息次数"""

    def with_accrued_interest(self) -> TfEvaluator:
        """计算应计利息"""

    def with_dirty_price(self) -> TfEvaluator:
        """计算债券全价"""

    def with_clean_price(self) -> TfEvaluator:
        """计算债券净价"""

    def with_duration(self) -> TfEvaluator:
        """计算修正久期"""

    def with_cf(self) -> TfEvaluator:
        """计算转换因子"""

    def with_basis_spread(self) -> TfEvaluator:
        """计算基差"""

    def with_deliver_accrued_interest(self) -> TfEvaluator:
        """
        计算国债期货交割应计利息

        国债期货交割应计利息=区间付息* (国债期货交割缴款日 - 国债期货交割前一付息日) / (国债期货交割下一付息日 - 国债期货交割前一付息日)

        按中金所发布公式, 计算结果四舍五入至小数点后7位
        """

    def with_future_dirty_price(self) -> TfEvaluator:
        """计算期货全价(发票价格)"""

    def with_remain_cp_to_deliver(self) -> TfEvaluator:
        """计算期间付息"""

    def with_deliver_cost(self) -> TfEvaluator:
        """
        计算交割成本

        交割成本=债券全价-期间��息
        """

    def with_f_b_spread(self) -> TfEvaluator:
        """计算期现价差"""

    def with_carry(self) -> TfEvaluator:
        """
        计算持有收益

        持有收益 = (交割日应计-交易日应计 + 期间付息) + 资金成本率*(加权平均期间付息-债券全价*剩余天数/365)
        """

    def with_net_basis_spread(self) -> TfEvaluator:
        """
        计算净基差

        净基差=基差-持有收益
        """

    def with_irr(self) -> TfEvaluator:
        """计算内部收益率IRR"""

    def with_future_ytm(self) -> TfEvaluator:
        """计算期货隐含收益率"""

    def calc_all(self) -> TfEvaluator:
        """计算所有指标"""

    def update(
        self,
        future_price: float | None = None,
        bond_ytm: float | None = None,
        date: date | None = None,
        future: Future | str | None = None,
        bond: Bond | str | int | None = None,
        capital_rate: float | None = None,
        reinvest_rate: float | None = None,
    ) -> TfEvaluator:
        """
        根据新的日期、债券和期货信息更新评估器

        此函数会根据输入的新信息更新评估器的各个字段,
        并根据变化情况决定是否保留原有的计算结果。
        当输入的参数为None时, 会沿用原始评估器中的值。
        """

    @property
    def accrued_interest(self) -> float:
        """应计利息, 如果未计算会自动计算后返回"""

    @property
    def deliver_accrued_interest(self) -> float:
        """国债期货交割应计利息, 如果未计算会自动计算后返回"""

    @property
    def cf(self) -> float:
        """转换因子, 如果未计算会自动计算后返回"""

    @property
    def dirty_price(self) -> float:
        """债券全价, 如果未计算会自动计算后返回"""

    @property
    def clean_price(self) -> float:
        """债券净价, 如果未计算会自动计算后返回"""

    @property
    def future_dirty_price(self) -> float:
        """期货全价, 如果未计算会自动计算后返回"""

    @property
    def deliver_cost(self) -> float:
        """交割成本, 如果未计算会自动计算后返回"""

    @property
    def basis_spread(self) -> float:
        """基差, 如果未计算会自动计算后返回"""

    @property
    def f_b_spread(self) -> float:
        """期现价差, 如果未计算会自动计算后返回"""

    @property
    def carry(self) -> float:
        """持有收益, 如果未计算会自动计算后返回"""

    @property
    def net_basis_spread(self) -> float:
        """净基差, 如果未计算会自动计算后返回"""

    @property
    def duration(self) -> float:
        """修正久期, 如果未计算会自动计算后返回"""

    @property
    def irr(self) -> float:
        """内部收益率, 如果未计算会自动计算后返回"""

    @property
    def deliver_date(self) -> date:
        """期货配对缴款日, 如果未计算会自动计算后返回"""

    @property
    def cp_dates(self) -> tuple[date, date] | None:
        """前一付息日和下一付息日, 如果未计算会自动计算后返回"""

    @property
    def deliver_cp_dates(self) -> tuple[date, date]:
        """期货交割日的前一付息日和下一付息日, 如果未计算会自动计算后返回"""

    @property
    def remain_cp_num(self) -> int | None:
        """债券剩余付息次数, 如果未计算会自动计算后返回"""

    @property
    def remain_cp_to_deliver(self) -> float:
        """到交割的期间付息, 如果未计算会自动计算后返回"""

    @property
    def remain_cp_to_deliver_wm(self) -> float:
        """加权平均到交割的期间付息, 如果未计算会自动计算后返回"""

    @property
    def future_ytm(self) -> float:
        """期货隐含收益率, 如果未计算会自动计算后返回"""

    def dv01(self) -> float:
        """计算DV01"""

    def future_dv01(
        self, ctd_bond: Bond | str | int | None = None, ctd_ytm: float = float("nan")
    ) -> float:
        """
        计算期货DV01

        Args:
            ctd_bond: CTD债券代码, 如果为None则使用当前债券
            ctd_ytm: CTD债券收益率
        """

    def neutral_cf(self, ctd_bond: Bond | str | int, ctd_ytm: float) -> float:
        """
        计算DV中性转换因子

        Args:
            ctd_bond: CTD债券代码
            ctd_ytm: CTD债券收益率
        """

    def neutral_net_basis_spread(
        self, ctd_bond: Bond | str | int, ctd_ytm: float
    ) -> float:
        """
        计算DV中性净基差

        dv中性净基差 = P - CF_Neutral * F - Carry

        Args:
            ctd_bond: CTD债券代码
            ctd_ytm: CTD债券收益率
        """



import datetime
import operator

from llvmlite import ir
from numba import types
from numba.core import cgutils
from numba.extending import (
    NativeValue,
    as_numba_type,
    box,
    lower_builtin,
    make_attribute_wrapper,
    models,
    overload,
    overload_attribute,
    overload_method,
    register_model,
    type_callable,
    typeof_impl,
    unbox,
)

from pybond.ffi import (
    datetime_add_duration,
    datetime_sub_duration,
    get_datetime_day,
    get_datetime_hour,
    get_datetime_minute,
    get_datetime_month,
    get_datetime_nanosecond,
    get_datetime_second,
    get_datetime_year,
    timestamp_nanos,
)

from .ir_utils import (
    ir_build_datetime,
    ir_isinstance,
    ir_local_timestamp_nanos,
    ir_timestamp_nanos,
)
from .nb_duration import DurationType
from .nb_time import Time


class DateTime:
    def __init__(self, val: int):
        self.ptr = 0
        self.val = int(val)

    @property
    def _pydt(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.val / 1e9)

    @property
    def time(self) -> datetime.time:
        return self._pydt.time

    @property
    def year(self) -> int:
        return self.pydt.year

    @property
    def month(self) -> int:
        return self._pydt.month

    @property
    def day(self) -> int:
        return self._pydt.day

    @property
    def hour(self) -> int:
        return self._pydt.hour

    @property
    def minute(self) -> int:
        return self._pydt.minute

    @property
    def second(self) -> int:
        return self._pydt.second

    @property
    def nanosecond(self) -> int:
        return int(self.val % 1e9)

    def __str__(self):
        return str(self._pydt)

    def __repr__(self):
        return repr(self._pydt)

    def __ge__(self, other):
        return self.val >= other.val

    def __gt__(self, other):
        return self.val > other.val

    def __lt__(self, other):
        return self.val < other.val

    def __le__(self, other):
        return self.val <= other.val

    def __eq__(self, other):
        return self.val == other.val

    def __ne__(self, other):
        return self.val != other.val


class DateTimeType(types.Type):
    def __init__(self):
        super().__init__(name="DateTime")


datetime_type = DateTimeType()
as_numba_type.register(datetime.datetime, datetime_type)


@typeof_impl.register(datetime.datetime)
@typeof_impl.register(DateTime)
def typeof_index(val, c):
    return datetime_type


@type_callable(DateTime)
def type_datetime(context):
    def typer(val):
        return datetime_type

    return typer


@register_model(DateTimeType)
class DateTimeModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("ptr", types.voidptr),
            ("val", types.int64),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


make_attribute_wrapper(DateTimeType, "ptr", "ptr")
make_attribute_wrapper(DateTimeType, "val", "val")


@lower_builtin(DateTime, types.int64)
@lower_builtin(DateTime, types.float64)
def impl_datetime_builder(context, builder, sig, args):
    typ = sig.return_type
    (val,) = args
    if isinstance(val.type, ir.DoubleType):
        val = builder.fptosi(val, ir.IntType(64))
    ptr = ir_build_datetime(val, builder)
    datetime_struct = cgutils.create_struct_proxy(typ)(context, builder)
    datetime_struct.ptr = ptr
    datetime_struct.val = val
    return datetime_struct._getvalue()


@unbox(DateTimeType)
def unbox_datetime(typ, obj, c):
    """
    Convert a Python datetime object or DateTime object to a native datetime structure.
    """
    # Get datetime.datetime type object
    datetime_type_c = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.datetime))

    # Check if object is instance of datetime.datetime
    is_datetime = ir_isinstance(c.pyapi, obj, datetime_type_c)
    c.pyapi.decref(datetime_type_c)

    val_ptr = cgutils.alloca_once(c.builder, ir.IntType(64))
    ptr_ptr = cgutils.alloca_once(c.builder, ir.PointerType(ir.IntType(8)))

    with c.builder.if_else(is_datetime) as (then, otherwise):
        with then:
            # Handle datetime.datetime
            ts_obj = c.pyapi.call_method(obj, "timestamp")
            val = c.pyapi.float_as_double(ts_obj)
            val = c.builder.fmul(
                val, ir.Constant(ir.DoubleType(), 1e9)
            )  # Convert to nanoseconds
            val = c.builder.fptosi(val, ir.IntType(64))  # Convert to int64
            c.pyapi.decref(ts_obj)
            ptr = ir_build_datetime(val, c.builder, from_utc=True)
            c.builder.store(ptr, ptr_ptr)
            c.builder.store(ir_timestamp_nanos(ptr, c.builder), val_ptr)

        with otherwise:
            # Handle DateTime object
            val_obj = c.pyapi.object_getattr_string(obj, "val")
            val = c.pyapi.long_as_longlong(val_obj)
            c.pyapi.decref(val_obj)
            ptr = ir_build_datetime(val, c.builder, from_utc=False)
            c.builder.store(ptr, ptr_ptr)
            c.builder.store(val, val_ptr)

    ptr = c.builder.load(ptr_ptr)
    val = c.builder.load(val_ptr)

    # Build the datetime struct with the computed timestamp value
    datetime_struct = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    datetime_struct.ptr = ptr
    datetime_struct.val = val

    # Check for errors
    is_error = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(datetime_struct._getvalue(), is_error=is_error)


@box(DateTimeType)
def box_datetime(typ, val, c):
    """
    Convert a native Datetime structure to python datetime.datetime.
    """
    dt = cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
    utc_timestamp_val = ir_local_timestamp_nanos(dt.ptr, c.builder)
    val = c.builder.sitofp(utc_timestamp_val, ir.DoubleType())  # Convert to double
    val = c.builder.fdiv(val, ir.Constant(ir.DoubleType(), 1e9))  # Convert to seconds
    # Call datetime.datetime.fromtimestamp()
    val_obj = c.pyapi.float_from_double(val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.datetime))
    res = c.pyapi.call_method(class_obj, "fromtimestamp", (val_obj,))
    c.pyapi.decref(val_obj)
    c.pyapi.decref(class_obj)
    return res


@overload_attribute(DateTimeType, "timestamp_nanos")
def get_timestamp_nanos_impl(dt):
    def getter(dt):
        return timestamp_nanos(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "year")
def get_year_impl(dt):
    def getter(dt):
        return get_datetime_year(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "month")
def get_month_impl(dt):
    def getter(dt):
        return get_datetime_month(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "day")
def get_day_impl(dt):
    def getter(dt):
        return get_datetime_day(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "hour")
def get_hour_impl(dt):
    def getter(dt):
        return get_datetime_hour(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "minute")
def get_minute_impl(dt):
    def getter(dt):
        return get_datetime_minute(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "second")
def get_second_impl(dt):
    def getter(dt):
        return get_datetime_second(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "nanosecond")
def get_nanosecond_impl(dt):
    def getter(dt):
        return get_datetime_nanosecond(dt.ptr)

    return getter


@overload_attribute(DateTimeType, "date")
def get_date_impl(dt):
    def getter(dt):
        return datetime.date(dt.year, dt.month, dt.day)

    return getter


@overload_attribute(DateTimeType, "time")
def get_time_impl(dt):
    def getter(dt):
        return Time(dt.hour, dt.minute, dt.second, dt.nanosecond)

    return getter


@overload_method(DateTimeType, "__str__")
def str_datetime(dt):
    """compile this method take a long time"""
    if not isinstance(dt, DateTimeType):
        return

    def impl(dt):
        return f"{str(dt.year).zfill(4)}-{str(dt.month).zfill(2)}-{str(dt.day).zfill(2)} {str(dt.hour).zfill(2)}:{str(dt.minute).zfill(2)}:{str(dt.second).zfill(2)}.{str(dt.nanosecond).zfill(9)[:6]}"

    return impl


@overload(operator.ge)
def impl_ge(dt, dt2):
    if not isinstance(dt, DateTimeType) or not isinstance(dt, DateTimeType):
        return

    def impl(dt, dt2):
        return dt.val >= dt2.val

    return impl


@overload(operator.gt)
def impl_gt(dt, dt2):
    if not isinstance(dt, DateTimeType) or not isinstance(dt, DateTimeType):
        return

    def impl(dt, dt2):
        return dt.val > dt2.val

    return impl


@overload(operator.lt)
def impl_lt(dt, dt2):
    if not isinstance(dt, DateTimeType) or not isinstance(dt, DateTimeType):
        return

    def impl(dt, dt2):
        return dt.val < dt2.val

    return impl


@overload(operator.le)
def impl_le(dt, dt2):
    if not isinstance(dt, DateTimeType) or not isinstance(dt, DateTimeType):
        return

    def impl(dt, dt2):
        return dt.val <= dt2.val

    return impl


@overload(operator.eq)
def impl_eq(dt, dt2):
    if not isinstance(dt, DateTimeType) or not isinstance(dt, DateTimeType):
        return

    def impl(dt, dt2):
        return dt.val == dt2.val

    return impl


@overload(operator.ne)
def impl_ne(dt, dt2):
    if not isinstance(dt, DateTimeType) or not isinstance(dt, DateTimeType):
        return

    def impl(dt, dt2):
        return dt.val != dt2.val

    return impl


@overload(operator.add)
def impl_datetime_add(val1, val2):
    if isinstance(val1, DateTimeType) and isinstance(val2, DurationType):

        def impl(val1, val2):
            return DateTime(datetime_add_duration(val1.val, val2.ptr))

        return impl
    elif isinstance(val1, DurationType) and isinstance(val2, DateTimeType):

        def impl(val1, val2):
            return DateTime(datetime_add_duration(val2.val, val1.ptr))

        return impl
    return


@overload(operator.sub)
def impl_datetime_sub(val1, val2):
    if isinstance(val1, DateTimeType) and isinstance(val2, DurationType):

        def impl(val1, val2):
            return DateTime(datetime_sub_duration(val1.val, val2.ptr))

        return impl
    return
